import {arrayToUtcDate, DateMarker, dateToUtcArray} from './marker.js'

export interface CalendarSystem {
  getMarkerYear(d: DateMarker): number
  getMarkerMonth(d: DateMarker): number
  getMarkerDay(d: DateMarker): number
  arrayToMarker(arr: number[]): DateMarker
  markerToArray(d: DateMarker): number[]
}

let calendarSystemClassMap = {}

export function registerCalendarSystem(name, theClass) {
  calendarSystemClassMap[name] = theClass
}

export function createCalendarSystem(name) {
  return new calendarSystemClassMap[name]()
}

class GregorianCalendarSystem implements CalendarSystem {
  getMarkerYear(d: DateMarker) {
    return d.getUTCFullYear()
  }

  getMarkerMonth(d: DateMarker) {
    return d.getUTCMonth()
  }

  getMarkerDay(d: DateMarker) {
    return d.getUTCDate()
  }

  arrayToMarker(arr) {
    return arrayToUtcDate(arr)
  }

  markerToArray(marker) {
    return dateToUtcArray(marker)
  }
}

registerCalendarSystem('gregory', GregorianCalendarSystem)

const JalaliDate = {
  g_days_in_month: [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
  j_days_in_month: [31, 31, 31, 31, 31, 31, 30, 30, 30, 30, 30, 29],
}

const jalaliToGregorian = function(j_y, j_m, j_d) {
  j_y = parseInt(j_y)
  j_m = parseInt(j_m)
  j_d = parseInt(j_d)
  let jy = j_y - 979
  let jm = j_m
  let jd = j_d - 1

  let j_day_no = 365 * jy + parseInt(String(jy / 33)) * 8 + parseInt(String((jy % 33 + 3) / 4))
  for (let i = 0; i < jm; ++i) j_day_no += JalaliDate.j_days_in_month[i]

  j_day_no += jd

  let g_day_no = j_day_no + 79

  let gy = 1600 + 400 * parseInt(String(g_day_no / 146097)) /* 146097 = 365*400 + 400/4 - 400/100 + 400/400 */
  g_day_no = g_day_no % 146097

  let leap = true
  if (g_day_no >= 36525) /* 36525 = 365*100 + 100/4 */
  {
    g_day_no--
    gy += 100 * parseInt(String(g_day_no / 36524)) /* 36524 = 365*100 + 100/4 - 100/100 */
    g_day_no = g_day_no % 36524

    if (g_day_no >= 365) g_day_no++
    else leap = false
  }

  gy += 4 * parseInt(String(g_day_no / 1461)) /* 1461 = 365*4 + 4/4 */
  g_day_no %= 1461

  if (g_day_no >= 366) {
    leap = false

    g_day_no--
    gy += parseInt(String(g_day_no / 365))
    g_day_no = g_day_no % 365
  }
  let i =0
  for (i = 0; g_day_no >= JalaliDate.g_days_in_month[i] + (i == 1 && leap ? 1 : 0); i++)
    g_day_no -= JalaliDate.g_days_in_month[i] + (i == 1 && leap ? 1 : 0)
  let gm = i + 1
  let gd = g_day_no + 1

  return [gy, gm-1, gd]
}


class PersianCalendarSystem implements CalendarSystem {






  getMarkerYear(d: DateMarker) {
    let fullYear = new Intl.DateTimeFormat('en-US-u-ca-persian', {year: 'numeric', timeZone : 'UTC',
      timeZoneName : 'short'}).format(d)
    let number = Number(fullYear.split(' ')[0])
    console.log('year :' +number)

    return number
  }

  getMarkerMonth(d: DateMarker) {
    let ret = new Intl.DateTimeFormat('en-US-u-ca-persian', {month: 'numeric' ,timeZone : 'UTC',
      timeZoneName : 'short'}).format(d)
    // console.log('month :' +ret)

    let number = Number(ret.split(',')[0])-1
    console.log('month :' +number)

    return number

  }

  getMarkerDay(d: DateMarker) {
    let day = new Intl.DateTimeFormat('en-US-u-ca-persian', {day: 'numeric', timeZone : 'UTC',
      timeZoneName : 'short'}).format(d)

    let ret = day.split(',')[0]
    console.log('day :' +ret)
    return Number(ret)
  }

  arrayToMarker(arr) {
    const newArr = jalaliToGregorian(arr[0],arr[1] || 0, arr[2] || 1)

    let date = arrayToUtcDate([...newArr, arr[3] || 0,arr[4] || 0,arr[5] || 0,arr[6] ||0 ])
    return new Date(date.getTime())
  }



  markerToArray(marker) {
    let ret = new Intl.DateTimeFormat('en-US-u-ca-persian', {
      timeZone: 'UTC',
      timeZoneName: 'short',
      year: 'numeric',
      month: 'numeric',
      day: 'numeric',
      hour: 'numeric',
      minute: 'numeric',
      second: 'numeric',
      hour12: false,
    }).format(marker).replace('AP', '').split(/[,:/\s]+/).map((it) => Number(it))
    const m = ret[0]
    const y = ret[2]
    const d=ret[1]
    ret[0]=y
    ret[1]=m
    ret[2]=d
    ret.splice(6)
    console.log('numbers: ')
    console.log(ret)

    return ret

    // const newArr = gregorianToJalali(ret[0],ret[1],ret[2])
    // return [...newArr, ret[3]||0,ret[4]||0,ret[5]||0,ret[6]||0]
  }
}

registerCalendarSystem('persian', PersianCalendarSystem)



const getMarkerDay=(d: DateMarker)=> {
  let day = new Intl.DateTimeFormat('en-US-u-ca-persian', {day: 'numeric', timeZone : 'UTC',
    timeZoneName : 'short'}).format(d)

  let ret = day.split(',')[0]
  console.log(ret)
  return Number(ret)
}


const getPersianOffset =() =>{
  let d=new Date()
  d.setDate(15)
  const pd= getMarkerDay(d)
  let offset = 15 - pd
  console.log('using pd offset : ' + offset)
  return offset
}

let offset =  getPersianOffset() *24*60*60*1000
class PersianLikeCalendarSystem implements CalendarSystem {



  getMarkerYear(d: DateMarker) {
    d = new Date(d.getTime()+offset)
    return d.getUTCFullYear()
  }

  getMarkerMonth(d: DateMarker) {
    d = new Date(d.getTime()+offset)
    return d.getUTCMonth()
  }

  getMarkerDay(d: DateMarker) {
    d = new Date(d.getTime()+offset)
    return d.getUTCDate()
  }

  arrayToMarker(arr) {
    return new Date(arrayToUtcDate(arr).getTime()-offset)
  }

  markerToArray(marker) {
    return dateToUtcArray(new Date(marker.getTime()+offset))
  }
}

registerCalendarSystem('persianLike', PersianLikeCalendarSystem)
